	
	----------------------------------------------------
	-- Mezzonic Cache Helper 1.01 by Fail US-Ragnaros --
	----------------------------------------------------

local ADDON_NAME,MezzonicCacheHelper = ...

--Create the Main Frame
MezzonicCacheHelper.Frame = MezzonicCacheHelper:FrameTemplate({
	hide = true,
	name = "MezzonicCacheSolverFrame",
	width = 350,
	height = 480,
	backdrop = {
		color = {0,0,0,.75},
		border = {1,1,1,1},
		edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border",
		edgeSize = 16,
		insets = {left=4,right=4,top=4,bottom=4},
	},
	text = {
		{
			name = "Description",
			text = string.format("|cffffd000%s v|r%s\n|cffffd000#|r1 |cffffd000Copy|r the |cffffd000Puzzle|r on the |cffffd000Grid|r above.\n|cffffd000#|r2 Click |cffffd000Solve|r\n|cffffd000#|r3 Click the |cffffd000Puzzle|r as shown in |cff00c000green|r.\n\n|cffffd000Created by|r %s",GetAddOnMetadata(ADDON_NAME,"Title"),GetAddOnMetadata(ADDON_NAME,"Version"),GetAddOnMetadata(ADDON_NAME,"Author")),
			point = "BOTTOMLEFT",
			xOff = 8,
			yOff = 8,
			size = 14,
			flag = "OUTLINE",
			horizontal = "LEFT",
			vertical = "TOP",
		},
		{
			name = "Percent",
			text = "%",
			point = "BOTTOMRIGHT",
			xOff = -8,
			yOff = 8,
			size = 14,
			flag = "OUTLINE",
			horizontal = "RIGHT",
			vertical = "TOP",
		},
	},
	movable = true,
	screen = true,
	script = {
		{
			type = "OnShow",
			func = function(self)
				self.ButtonSolve:Enable()
				self.ButtonSolve.Text:SetTextColor(1,1,1,1)
				self.ButtonSolve:SetBackdropBorderColor(0,0,0,1)
				self.Input = {{0,0,0,0,0},{0,0,0,0,0},{0,0,0,0,0},{0,0,0,0,0},{0,0,0,0,0}}
				self:Update(self.Input,true)
				self:ClearAllPoints()
				self:SetPoint(MezzonicCacheHelperDB.point,UIParent,MezzonicCacheHelperDB.relativePoint,MezzonicCacheHelperDB.xOff,MezzonicCacheHelperDB.yOff)
				self:SetScale(MezzonicCacheHelperDB.scale)
				self.Scale:SetText(MezzonicCacheHelperDB.scale*100)
			end,
		},
		{
			type = "OnMouseUp",
			func = function(self,button)
				if button ~= "LeftButton" then return end
				local point,_,relativePoint,xOff,yOff = self:GetPoint()
				MezzonicCacheHelperDB.point = point
				MezzonicCacheHelperDB.relativePoint = relativePoint
				MezzonicCacheHelperDB.xOff = xOff
				MezzonicCacheHelperDB.yOff = yOff
			end,
		},
	},
	event = {
		{
			event = "ADDON_LOADED",
			func = function(self,event,addonName)
				if addonName ~= ADDON_NAME then return end
				self:UnregisterEvent(event)
				local defaultDB = {scale = 1,point = "RIGHT",relativePoint = "RIGHT",xOff = -371,yOff = 40}
				if MezzonicCacheHelperDB == nil then
					MezzonicCacheHelperDB = defaultDB
				else
					for key,value in pairs(defaultDB) do
						if MezzonicCacheHelperDB[key] == nil then
							MezzonicCacheHelperDB[key] = value
						end
					end
				end
			end,
		},
		{
			event = {"PLAYER_ENTERING_WORLD","ZONE_CHANGED"},
			func = function(self)
				local info = C_Map.GetMapInfo(C_Map.GetBestMapForUnit("player"))
				--check if we are on Zereth Mortis
				if info.mapID == 1970 then
					self:RegisterUnitEvent("UNIT_AURA","player")
					self:Check()
				else --if not unregister UNIT_AURA
					self:UnregisterEvent("UNIT_AURA")
				end
			end,
		},
		{
			event = "UNIT_AURA",
			unregister = true,
			func = function(self,_,unitID)
				if unitID ~= "player" then return end
				self:Check()
			end,
		},
	},
})

--Simple Scale
MezzonicCacheHelper.Frame.Scale = MezzonicCacheHelper:FrameTemplate({
	type = "EDITBOX",
	parent = MezzonicCacheHelper.Frame,
	point = "BOTTOMRIGHT",
	xOff = -15,
	yOff = -9,
	width = 64,
	height = 48,
	numeric = true,
	horizontal = "RIGHT",
	insets = {8,8,8,8},
	script = {
		{
			type = "OnEscapePressed",
			func = function(self)
				self:ClearFocus()
			end,
		},
		{
			type = "OnEnterPressed",
			func = function(self)
				self:ClearFocus()
			end,
		},
		{
			type = "OnEditFocusLost",
			func = function(self)
				local value = math.floor(math.min(math.max(self:GetNumber(),75),150))
				self:SetText(value)
				self:GetParent():SetScale(value/100)
				MezzonicCacheHelperDB.scale = value/100
			end,
		},
	},
})
MezzonicCacheHelper.Frame.Percent:ClearAllPoints()
MezzonicCacheHelper.Frame.Percent:SetPoint("LEFT",MezzonicCacheHelper.Frame.Scale,"RIGHT",-8,0)

--Create the Grid
MezzonicCacheHelper.Frame.Grid = {{},{},{},{},{}}
MezzonicCacheHelper.Frame.Input = {{0,0,0,0,0},{0,0,0,0,0},{0,0,0,0,0},{0,0,0,0,0},{0,0,0,0,0}}
for index=0,24 do
	MezzonicCacheHelper.Frame["ButtonGrid"..index+1] = MezzonicCacheHelper:FrameTemplate({
		type = "BUTTON",
		parent = MezzonicCacheHelper.Frame,
		point = "TOPLEFT",
		xOff = ((index%5)*64)+16,
		yOff = (math.floor(index/5)*-64)-16,
		width = 64,
		height = 64,
		backdrop = {
			color = {0,0,0,0},
			border = {1,1,1,1},
		},
		script = {
			{
				type = "OnClick",
				func = function(self)
					local parent = self:GetParent()
					parent.Input[self.row][self.col] = (parent.Input[self.row][self.col]+1)%2
					if parent.Input[self.row][self.col] == 1 then
						self:SetBackdropColor(1,1,1,1)
						self:SetBackdropBorderColor(0,0,0,1)
					else
						self:SetBackdropColor(0,0,0,1)
						self:SetBackdropBorderColor(1,1,1,1)
					end
				end,
			},
		},
	})
	MezzonicCacheHelper.Frame["ButtonGrid"..index+1].row = math.floor(index/5)+1
	MezzonicCacheHelper.Frame["ButtonGrid"..index+1].col = (index%5)+1
	MezzonicCacheHelper.Frame.Grid[math.floor(index/5)+1][(index%5)+1] = MezzonicCacheHelper.Frame["ButtonGrid"..index+1]
end

--Solve Button
MezzonicCacheHelper.Frame.ButtonSolve = MezzonicCacheHelper:FrameTemplate({
	type = "BUTTON",
	parent = MezzonicCacheHelper.Frame,
	point = "TOPLEFT",
	relative = MezzonicCacheHelper.Frame.ButtonGrid21,
	relativePoint = "BOTTOMLEFT",
	yOff = -8,
	width = 152,
	height = 24,
	backdrop = {
		color = {.1,.1,.1,1},
		border = {0,0,0,1},
	},
	text = {
		{
			text = "Solve",
			size = 16,
			flag = "OUTLINE",
		},
	},
	script = {
		{
			type = "OnEnter",
			func = function(self)
				self:SetBackdropBorderColor(0,.75,0,1)
				self.Text:SetTextColor(0,.75,0,1)
			end,
		},
		{
			type = "OnLeave",
			func = function(self)
				self:SetBackdropBorderColor(0,0,0,1)
				self.Text:SetTextColor(1,1,1,1)
			end,
		},
		{
			type = "OnClick",
			func = function(self)
				PlaySound(687,"Master")
				local parent = self:GetParent()
				self:Disable()
				self:SetBackdropBorderColor(0,.75,0,1)
				self.Text:SetTextColor(0,.75,0,1)
				--for the most optimal solution we need to check the grid upside down too
				local upside = {{},{},{},{},{}}
				for row=1,5 do
					for col=1,5 do
						upside[6-row][col] = parent.Input[row][col]+0 --inverting the rows
					end
				end
				local output1,moves1,unsolvable1 = parent:Solve(parent.Input) --original
				local output2,moves2,unsolvable2 = parent:Solve(upside) --inversed
				if not unsolvable2 and moves2 < moves1 then --if the inversed is solvable and has less moves then we prefer it
					for row=1,5 do
						for col=1,5 do
							upside[6-row][col] = output2[row][col]+0
						end
					end
					parent:Update(upside,false)
					return
				end
				parent:Update(output1,false,unsolvable1)
			end,
		},
	},
})

--Reset Button
MezzonicCacheHelper.Frame.ButtonReset = MezzonicCacheHelper:FrameTemplate({
	type = "BUTTON",
	parent = MezzonicCacheHelper.Frame,
	point = "TOPLEFT",
	relative = MezzonicCacheHelper.Frame.ButtonSolve,
	relativePoint = "TOPRIGHT",
	xOff = 16,
	width = 152,
	height = 24,
	backdrop = {
		color = {.1,.1,.1,1},
		border = {0,0,0,1},
	},
	text = {
		{
			text = "Reset",
			size = 16,
			flag = "OUTLINE",
		},
	},
	script = {
		{
			type = "OnEnter",
			func = function(self)
				self:SetBackdropBorderColor(.75,0,0,1)
				self.Text:SetTextColor(.75,0,0,1)
			end,
		},
		{
			type = "OnLeave",
			func = function(self)
				self:SetBackdropBorderColor(0,0,0,1)
				self.Text:SetTextColor(1,1,1,1)
			end,
		},
		{
			type = "OnClick",
			func = function(self)
				PlaySound(687,"Master")
				local parent = self:GetParent()
				parent.ButtonSolve:Enable()
				parent.ButtonSolve:SetBackdropBorderColor(0,0,0,1)
				parent.ButtonSolve.Text:SetTextColor(1,1,1,1)
				parent.Input = {{0,0,0,0,0},{0,0,0,0,0},{0,0,0,0,0},{0,0,0,0,0},{0,0,0,0,0}}
				parent:Update(parent.Input,true)
			end,
		},
	},
})

function MezzonicCacheHelper.Frame:Check()
	--check if we are on a Mezzonic Protolock
	for index=1,255 do
		local _,_,_,_,_,_,_,_,_,spellID = UnitAura("player",index)
		if not spellID then break end
		if spellID == 351405 or spellID == 366042 or spellID == 366106 then
			self:Show()
			return
		end
	end
	self:Hide()
end

function MezzonicCacheHelper.Frame:Toggle(grid,row,col)
	--we toggle the lights from the row/col and adjacent cells
	if grid[row-1] and grid[row-1][col] then grid[row-1][col] = (grid[row-1][col]+1)%2 end
    if grid[row] and grid[row][col+1] then grid[row][col+1] = (grid[row][col+1]+1)%2 end
    if grid[row] and grid[row][col-1] then grid[row][col-1] = (grid[row][col-1]+1)%2 end
    if grid[row+1] and grid[row+1][col] then grid[row+1][col] = (grid[row+1][col]+1)%2 end
    if grid[row]and grid[row][col] then grid[row][col] = (grid[row][col]+1)%2 end
end

--this will not always show the most efficient way of solving the puzzle
--but it will show a solution that will be efficient most of the time
function MezzonicCacheHelper.Frame:Solve(input)
	--we need to simulate solving the grid using the Chase The Lights method
    local output = {{0,0,0,0,0},{0,0,0,0,0},{0,0,0,0,0},{0,0,0,0,0},{0,0,0,0,0}} --what moves we need to make
	local check = {{},{},{},{},{}} --we need to do a double-check after we are over	
	for row=1,5 do
		for col=1,5 do
			check[row][col] = input[row][col]+0 --creating a copy of the grid to double-check
		end
	end
	local last = "" --we need to save the last row
	for iteration=1,2 do --we go across the grid twice
		for row=1,5 do
			for col=1,5 do
				if row == 5 then
					--row 5 has special rules
					if iteration == 2 then break end --if this is the second time then we skip this
					last = last..(input[row][col] == 1 and col or "") --saving which lights are ON as a string
					if col == 5 then
						--we can then check which combination is ON, this is the special rule
						--we toggle the first row according to which combination we got
						if last == "123" then
							self:Toggle(input,1,2)
						elseif last == "1245" then
							self:Toggle(input,1,3)
						elseif last == "134" then
							self:Toggle(input,1,5)
						elseif last == "15" then
							self:Toggle(input,1,1)
							self:Toggle(input,1,2)
						elseif last == "235" then
							self:Toggle(input,1,1)
						elseif last == "24" then
							self:Toggle(input,1,1)
							self:Toggle(input,1,4)
						elseif last == "345" then
							self:Toggle(input,1,4)
						end
					end
				elseif input[row][col] == 1 then
					--if a light is ON then we do two things
					--we need to toggle the light on the next row directly below that one
					--we also save this move to give the output at the end
					if input[row+1] and input[row+1][col] then self:Toggle(input,row+1,col) end
					if output[row+1] and output[row+1][col] then output[row+1][col] = (output[row+1][col]+1)%2 end
				end
			end
		end
	end
	--we then use our output to toggle the lights on the double-check grid
	local moves = 0
	for row=1,5 do
		for col=1,5 do
			if output[row][col] == 1 then
				moves = moves+1
				self:Toggle(check,row,col)
			end
		end
	end
	--then we check row 2 of the double-check grid for any ON lights
	for col=1,5 do
		if check[2][col] == 1 then --we find a ON light
			moves = moves+1
			output[1][col] = 1 --we add to the output the light above it to complete the puzzle
		end
	end
	--we complete the double-check grid
	for col=1,5 do
		if output[1][col] == 1 then
			self:Toggle(check,1,col)
		end
	end
	--after all of this we check if there are still any ON light
	for row=1,5 do
		for col=1,5 do
			if check[row][col] == 1 then --if we find any ON light it means the grid has no solution
				return check,99,true
			end
		end
	end
	return output,moves
end

function MezzonicCacheHelper.Frame:Update(grid,enable,unsolvable)
	if unsolvable then --if the grid cannot be solved then we turn everything red
		self.ButtonSolve:SetBackdropBorderColor(.75,0,0,1)
		self.ButtonSolve.Text:SetTextColor(.75,0,0,1)
	end
	--update the grid according to the input
	for row=1,5 do
		for col=1,5 do
			self.Grid[row][col]:EnableMouse(enable)
			if grid[row][col] == 1 then
				if unsolvable then --if the grid cannot be solved then we turn everything red
					self.Grid[row][col]:SetBackdropColor(.75,0,0,1)
					self.Grid[row][col]:SetBackdropBorderColor(1,1,1,1)
				else --else the grid is green
					self.Grid[row][col]:SetBackdropColor(0,.75,0,1)
					self.Grid[row][col]:SetBackdropBorderColor(1,1,1,1)
				end
			else
				self.Grid[row][col]:SetBackdropColor(0,0,0,1)
				self.Grid[row][col]:SetBackdropBorderColor(1,1,1,1)
			end
		end
	end
end
